import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseWidget


def plot_agreement_matrix(sorting_comparison, ordered=True, count_text=True,
                    unit_ticks=True, ax=None, figure=None):
    """
    Plots sorting comparison confusion matrix.

    Parameters
    ----------
    sorting_comparison: GroundTruthComparison or SymmetricSortingComparison
        The sorting comparison object.
        Symetric or not.
    ordered: bool
        Order units with best agreement scores.
        This enable to see agreement on a diagonal.
    count_text: bool
        If True counts are displayed as text
    unit_ticks: bool
        If True unit tick labels are displayed
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: ConfusionMatrixWidget
        The output widget
    """
    W = AgreementMatrixWidget(
        sorting_comparison=sorting_comparison,
        ordered=ordered,
        count_text=count_text,
        unit_ticks=unit_ticks,
        figure=figure,
        ax=ax,
    )
    W.plot()
    return W


class AgreementMatrixWidget(BaseWidget):
    def __init__(self, *, sorting_comparison, ordered=True, count_text=True, unit_ticks=True,
                 figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._sc = sorting_comparison
        self._ordered = ordered
        self._count_text = count_text
        self._unit_ticks = unit_ticks
        self.name = 'ConfusionMatrix'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        # a dataframe
        if self._ordered:
            scores = self._sc.get_ordered_agreement_scores()
        else:
            scores = self._sc.agreement_scores

        N1 = scores.shape[0]
        N2 = scores.shape[1]
        
        unit_ids1 = scores.index.values
        unit_ids2 = scores.columns.values

        # Using matshow here just because it sets the ticks up nicely. imshow is faster.
        self.ax.matshow(scores.values, cmap='Greens')

        if self._count_text:
            for i, u1 in enumerate(unit_ids1):
                u2 = self._sc.best_match_12[u1]
                if u2 != -1:
                    j = np.where(unit_ids2==u2)[0][0]
                    
                    self.ax.text(j, i, '{:0.2f}'.format(scores.at[u1, u2]),
                                        ha='center', va='center', color='white')
        
        # Major ticks
        self.ax.set_xticks(np.arange(0, N2))
        self.ax.set_yticks(np.arange(0, N1))
        self.ax.xaxis.tick_bottom()
        
        # Labels for major ticks
        if self._unit_ticks:
            self.ax.set_yticklabels(scores.index, fontsize=12)
            self.ax.set_xticklabels(scores.columns, fontsize=12)

        self.ax.set_xlabel(self._sc.name_list[1], fontsize=20)
        self.ax.set_ylabel(self._sc.name_list[0], fontsize=20)
        
        #~ self.ax.set_ylim(-0.5, N1-0.5)
        #~ self.ax.set_xlim(N2-0.5, -0.5, )

        self.ax.set_xlim(-0.5, N2-0.5)
        self.ax.set_ylim(N1-0.5, -0.5, )
        
        




import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseMultiWidget


def plot_amplitudes_distribution(recording, sorting, unit_ids=None, max_spikes_per_unit=100,
                                 figure=None, ax=None):
    """
    Plots waveform amplitudes distribution.

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor object
    sorting: SortingExtractor
        The sorting extractor object
    unit_ids: list
        List of unit ids
    max_spikes_per_unit: int
        Maximum number of spikes to display per unit.
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: AmplitudeDistributionWidget
        The output widget
    """
    W = AmplitudeDistributionWidget(
        sorting=sorting,
        recording=recording,
        unit_ids=unit_ids,
        max_spikes_per_unit=max_spikes_per_unit,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


def plot_amplitudes_timeseries(recording, sorting, unit_ids=None, max_spikes_per_unit=100,
                               figure=None, ax=None):
    """
    Plots waveform amplitudes timeseries.

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor object
    sorting: SortingExtractor
        The sorting extractor object
    unit_ids: list
        List of unit ids
    max_spikes_per_unit: int
        Maximum number of spikes to display per unit.
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: AmplitudeTimeseriesWidget
        The output widget
    """
    W = AmplitudeTimeseriesWidget(
        sorting=sorting,
        recording=recording,
        unit_ids=unit_ids,
        max_spikes_per_unit=max_spikes_per_unit,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class AmplitudeBaseWidget(BaseMultiWidget):
    def __init__(self, recording, sorting, max_spikes_per_unit=100, figure=None, ax=None):
        BaseMultiWidget.__init__(self, figure, ax)
        self._sorting = sorting
        self._recording = recording
        self._max_spikes_per_unit = max_spikes_per_unit

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        raise NotImplementedError

    def compute_amps(self, *, times):
        amps = np.zeros(len(times))
        times_ind = (times * self._recording.get_sampling_frequency()).astype(int)
        for i, t in enumerate(times_ind):
            amp = self._recording.get_traces(start_frame=t-1, end_frame=t+1)
            amps[i] = np.min(amp)
        return amps


class AmplitudeTimeseriesWidget(AmplitudeBaseWidget):
    def __init__(self, *, recording, sorting, unit_ids=None, max_spikes_per_unit=100, figure=None, ax=None):
        AmplitudeBaseWidget.__init__(self, recording, sorting, max_spikes_per_unit, figure, ax)
        self._unit_ids = unit_ids
        self.name = 'AmplitudeTimeseries'

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        list_amps = []
        all_amps = np.array([])
        for unit in units:
            times = self._sorting.get_unit_spike_train(unit_id=unit) / \
                    float(self._recording.get_sampling_frequency())
            if len(times) > self._max_spikes_per_unit:
                times = times[np.random.permutation(len(times))[:self._max_spikes_per_unit]]
            amps = self.compute_amps(times=times)
            item = dict(
                title='Unit {}'.format(int(unit)),
                times=times,
                amps=amps
            )
            list_amps.append(item)
            all_amps = np.concatenate((all_amps, amps))
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            ylim = [np.min(all_amps) - 10, np.max(all_amps) + 10]
            self._plot_amp_time_multi(list_amps, ylim=ylim)

    def _plot_amp_time_multi(self, list_isi, *, ncols=5, ylim=None, **kwargs):
        if len(list_isi) < ncols:
            ncols = len(list_isi)
        nrows = np.ceil(len(list_isi) / ncols)
        for i, item in enumerate(list_isi):
            ax = self.get_tiled_ax(i, nrows, ncols, hspace=0.7, wspace=1)
            if i == 0 or i == ncols:
                ylab = True
            else:
                ylab = False
            _plot_amp_time(**item, **kwargs, ax=ax, ylim=ylim, ylab=ylab)


class AmplitudeDistributionWidget(AmplitudeBaseWidget):
    def __init__(self, *, recording, sorting, unit_ids=None, max_spikes_per_unit=100, figure=None, ax=None):
        AmplitudeBaseWidget.__init__(self, recording, sorting, max_spikes_per_unit, figure, ax)
        self._unit_ids = unit_ids
        self.name = 'AmplitudeDistribution'

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        list_amps = []
        all_amps = np.array([])
        for unit in units:
            times = self._sorting.get_unit_spike_train(unit_id=unit) / \
                    float(self._recording.get_sampling_frequency())
            if len(times) > self._max_spikes_per_unit:
                times = times[np.random.permutation(len(times))[:self._max_spikes_per_unit]]
            amps = self.compute_amps(times=times)
            item = dict(
                title='Unit {}'.format(int(unit)),
                times=times,
                amps=amps
            )
            list_amps.append(item)
            all_amps = np.concatenate((all_amps, amps))
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_amp_hist_multi(list_amps)

    def _plot_amp_hist_multi(self, list_isi, *, ncols=5, **kwargs):
        if len(list_isi) < ncols:
            ncols = len(list_isi)
        nrows = np.ceil(len(list_isi) / ncols)
        for i, item in enumerate(list_isi):
            ax = self.get_tiled_ax(i, nrows, ncols, hspace=0.7)
            _plot_amp_dist(**item, **kwargs, ax=ax)


def roundup(x, num=5):
    return int(np.ceil(x / num)) * num


def rounddown(x, num=5):
    return int(np.floor(x / num)) * num


def _plot_amp_dist(*, times, amps, ax, title=''):
    std_amps = np.std(amps)
    idxs = (amps > np.median(amps) - 4 * std_amps) & (amps < np.median(amps) + 4 * std_amps)
    amps = amps[idxs]
    h, b, _ = ax.hist(amps, density=True, color='gray', bins=10)
    ax.set_yticks([])
    ax.set_xticks([rounddown(np.mean(amps) - 2 * np.std(amps)),
                   roundup(np.mean(amps) + 2 * np.std(amps))])
    ax.set_xlabel('amp')
    if title:
        ax.set_title(title, color='gray')


def _plot_amp_time(*, times, amps, ax, title='', ylim=None, ylab=False):
    std_amps = np.std(amps)
    idxs = (amps > np.median(amps) - 4 * std_amps) & (amps < np.median(amps) + 4 * std_amps)
    amps = amps[idxs]
    times = times[idxs]
    ax.plot(times, amps, '*', color='gray', alpha=0.3)
    if ylab:
        ax.set_ylabel('amp')
    ax.set_xlabel('time (s)')
    ax.set_xticks([])
    if ylim is not None:
        ax.set_ylim(ylim)
    if title:
        ax.set_title(title, color='gray')



import numpy as np
from spikewidgets.widgets.basewidget import BaseWidget


def plot_spectrum(recording, channels=None, trange=None, freqrange=None, color_groups=False, color='steelblue',
                  nfft=256, figure=None, ax=None):
    """
    Plots electrode geometry.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    channels: list
        The channels to show
    trange: list
        List with start time and end time
    freqrange: list
        List with start frequency and end frequency
    color_groups: bool
        If True groups are plotted with different colors
    color: matplotlib color
        The color to be used
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: TimeseriesWidget
        The output widget
    """
    W = SpectrumWidget(
        recording=recording,
        channels=channels,
        trange=trange,
        freqrange=freqrange,
        color_groups=color_groups,
        color=color,
        nfft=nfft,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


def plot_spectrogram(recording, channel, trange=None, freqrange=None, cmap='viridis',
                     nfft=256, figure=None, ax=None):
    """
    Plots electrode geometry.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    channel: int
        The channel to plot spectrogram of
    trange: list
        List with start time and end time
    freqrange: list
        List with start frequency and end frequency
    cmap: matplotlib colorma
        The colormap to be used
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: TimeseriesWidget
        The output widget
    """
    W = SpectrogramWidget(
        recording=recording,
        channel=channel,
        trange=trange,
        freqrange=freqrange,
        cmap=cmap,
        nfft=nfft,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class SpectrumWidget(BaseWidget):
    def __init__(self, *, recording, channels=None, trange=None, freqrange=None,
                 color_groups=False, color=None, nfft=256, figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._recording = recording
        self._samplerate = recording.get_sampling_frequency()
        self._ax = ax
        self._color_groups = color_groups
        self._trange = trange
        self._frange = freqrange
        self._color = color
        self._nfft = nfft

        if channels is None:
            self._channels = self._recording.get_channel_ids()
        else:
            self._channels = channels

        if color_groups:
            self._colors = []
            self._group_color_map = {}
            all_groups = recording.get_channel_groups()
            groups = np.unique(all_groups)
            N = len(groups)
            import colorsys
            HSV_tuples = [(x * 1.0 / N, 0.5, 0.5) for x in range(N)]
            self._colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples))
            color_idx = 0
            for group in groups:
                self._group_color_map[group] = color_idx
                color_idx += 1
        self.name = 'Spectrum'

    def plot(self):
        if self._trange is not None:
            start_frame = int(self._trange[0] * self._recording.get_sampling_frequency())
            end_frame = int(self._trange[1] * self._recording.get_sampling_frequency())
        else:
            start_frame = 0
            end_frame = self._recording.get_num_frames()

        traces = self._recording.get_traces(channel_ids=self._channels, start_frame=start_frame, end_frame=end_frame)

        for (t, ch) in zip(traces, self._channels):
            if self._color_groups:
                group = self._recording.get_channel_groups(channel_ids=[ch])[0]
                group_color_idx = self._group_color_map[group]
                self.ax.psd(t, Fs=self._samplerate, color=self._colors[group_color_idx], NFFT=self._nfft)
            else:
                self.ax.psd(t, Fs=self._samplerate, NFFT=self._nfft)

        if self._frange is not None:
            self.ax.set_xlim(self._frange)


class SpectrogramWidget(BaseWidget):
    def __init__(self, *, recording, channel, trange=None, freqrange=None,
                 cmap=None, nfft=256, figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._recording = recording
        self._samplerate = recording.get_sampling_frequency()
        self._ax = ax
        self._trange = trange
        self._frange = freqrange
        self._cmap = cmap
        self._nfft = nfft

        assert channel is not None, "Specify which 'channel' to use"
        self._channel = channel
        self.name = 'Spectrocram'

    def plot(self):
        if self._trange is not None:
            start_frame = int(self._trange[0] * self._recording.get_sampling_frequency())
            end_frame = int(self._trange[1] * self._recording.get_sampling_frequency())
        else:
            start_frame = 0
            end_frame = self._recording.get_num_frames()

        traces = self._recording.get_traces(channel_ids=[self._channel], start_frame=start_frame, end_frame=end_frame)

        self.ax.specgram(traces[0], Fs=self._recording.get_sampling_frequency(), scale='dB', cmap=self._cmap,
                         NFFT=self._nfft)

        if self._frange is not None:
            self.ax.set_ylim(self._frange)

        if self._trange is not None:
            l_xticks = len(self.ax.get_xticks())
            self.ax.set_xticklabels(np.linspace(self._trange[0], self._trange[1], l_xticks))


import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np


class BaseWidget:
    def __init__(self, figure=None, ax=None):
        if figure is None and ax is None:
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
        elif ax is None:
            self.figure = figure
            self.ax = self.figure.add_subplot(111)
        else:
            self.figure = ax.get_figure()
            self.ax = ax
        self.name = None
    
    def get_figure(self):
        return self.figure
    
    def get_ax(self):
        return self.ax
    
    def get_name(self):
        return self.name

class BaseMultiWidget:
    def __init__(self, figure=None, ax=None):
        if figure is None and ax is None:
            self.figure = plt.figure()
            self.ax = self.figure.add_subplot(111)
        elif ax is None:
            self.figure = figure
            self.ax = self.figure.add_subplot(111)
        else:
            self.figure = ax.get_figure()
            self.ax = ax
        self.axes = []
        self._gs = None
        self.ax.axis('off')
        self.name = None

    def get_tiled_ax(self, i, nrows, ncols, hspace=0.3, wspace=0.3, is_diag=False):
        if self._gs is None:
            self._gs = gridspec.GridSpecFromSubplotSpec(int(nrows), int(ncols), subplot_spec=self.ax,
                                                        hspace=hspace, wspace=wspace)
        r = int(i // ncols)
        c = int(np.mod(i, ncols))
        gs_sel = self._gs[r, c]
        ax = self.figure.add_subplot(gs_sel)
        self.axes.append(ax)
        if r == c:
            diag = True
        else:
            diag = False
        if is_diag:
            return ax, diag
        else:
            return ax



import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseMultiWidget
import spiketoolkit as st


def plot_pca_features(recording, sorting, unit_ids=None, max_spikes_per_unit=100, nproj=4, colormap=None,
                      figure=None, ax=None):
    """
    Plots unit PCA features on best projections.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object.
    sorting: SortingExtractor
        The sorting extractor object.
    unit_ids: list
        List of unit ids.
    max_spikes_per_unit: int
        Maximum number of spikes to display per unit.
    nproj: int
        Number of best projections to display.
    colormap: matplotlib colormap
        The colormap to be used. If not given default is used.
    figure: matplotlib figure
        The figure to be used. If not given a figure is created.
    ax: matplotlib axis
        The axis to be used. If not given an axis is created.

    Returns
    -------
    W: PCAWidget
        The output widget
    """
    W = PCAWidget(
        sorting=sorting,
        recording=recording,
        unit_ids=unit_ids,
        max_spikes_per_unit=max_spikes_per_unit,
        nproj=nproj,
        colormap=colormap,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class PCAWidget(BaseMultiWidget):
    def __init__(self, *, recording, sorting, unit_ids=None, max_spikes_per_unit=100, nproj=4, colormap=None,
                 figure=None, ax=None, save_as_features=False, save_waveforms_as_features=False):
        BaseMultiWidget.__init__(self, figure, ax)
        self._sorting = sorting
        self._recording = recording
        self._unit_ids = unit_ids
        self._nproj = nproj
        self._max_spikes_per_unit = max_spikes_per_unit
        self._pca_scores = None
        self._colormap = colormap
        self._save_as_features = save_as_features
        self._save_waveforms_as_features = save_waveforms_as_features
        self.name = 'Feature'

    def _compute_pca(self):
        self._pca_scores = st.postprocessing.compute_unit_pca_scores(recording=self._recording,
                                                                     sorting=self._sorting,
                                                                     by_electrode=True,
                                                                     max_spikes_per_unit=self._max_spikes_per_unit,
                                                                     save_as_features=self._save_as_features,
                                                                     save_waveforms_as_features=self._save_waveforms_as_features)

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        self._units = units

        if self._pca_scores is None:
            self._compute_pca()

        # find projections with best separation
        n_pc = self._pca_scores[0].shape[2]
        n_ch = self._pca_scores[0].shape[1]

        distances = []
        proj = []
        for ch1 in range(n_ch):
            for pc1 in range(n_pc):
                for ch2 in range(n_ch):
                    for pc2 in range(n_pc):
                        if ch1 != ch2 or pc1 != pc2:
                            dist = self.compute_cluster_average_distance(
                                pc1, ch1, pc2, ch2)
                            if [ch1, pc1, ch2, pc2] not in proj and [
                                    ch2, pc2, ch1, pc1] not in proj:
                                distances.append(dist)
                                proj.append([ch1, pc1, ch2, pc2])

        list_best_proj = np.array(
            proj)[np.argsort(distances)[::-1][:self._nproj]]
        self._plot_proj_multi(list_best_proj)

    def compute_cluster_average_distance(self, pc1, ch1, pc2, ch2):
        centroids = np.zeros((len(self._pca_scores), 2))
        for i, pcs in enumerate(self._pca_scores):
            centroids[i, 0] = np.median(pcs[:, ch1, pc1], axis=0)
            centroids[i, 1] = np.median(pcs[:, ch2, pc2], axis=0)

        dist = []
        for i, c1 in enumerate(centroids):
            for j, c2 in enumerate(centroids):
                if i > j:
                    dist.append(np.linalg.norm(c2 - c1))

        return np.mean(dist)

    def _plot_proj_multi(self, best_proj, ncols=5):
        if len(best_proj) < ncols:
            ncols = len(best_proj)
        nrows = np.ceil(len(best_proj) / ncols)
        for i, bp in enumerate(best_proj):
            ax = self.get_tiled_ax(i, nrows, ncols, hspace=0.3)
            self._plot_proj(proj=bp, ax=ax)

    def _plot_proj(self, *, proj, ax, title=''):
        ch1, pc1, ch2, pc2 = proj
        if self._colormap is not None:
            cm = plt.get_cmap(self._colormap)
            colors = [cm(i / len(self._pca_scores))
                      for i in np.arange(len(self._pca_scores))]
        else:
            colors = [None] * len(self._pca_scores)
        for i, pc in enumerate(self._pca_scores):
            if self._sorting.get_unit_ids()[i] in self._units:
                ax.plot(pc[:, ch1, pc1], pc[:, ch2, pc2],
                        '*', color=colors[i], alpha=0.3)
        ax.set_yticks([])
        ax.set_xticks([])
        ax.set_xlabel('ch {} - pc {}'.format(ch1, pc1))
        ax.set_ylabel('ch {} - pc {}'.format(ch2, pc2))
        if title:
            ax.set_title(title, color='gray')



import numpy as np
from spikewidgets.widgets.basewidget import BaseWidget


def plot_electrode_geometry(recording, markersize=20, marker='o', figure=None, ax=None):
    """
    Plots electrode geometry.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    markersize: int
        The size of the marker for the electrodes
    marker: str
        The matplotlib marker to use (default 'o')
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: UnitWaveformsWidget
        The output widget
    """
    W = ElectrodeGeometryWidget(
        recording=recording,
        markersize=markersize,
        marker=marker,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class ElectrodeGeometryWidget(BaseWidget):
    def __init__(self, *, recording, markersize=10, marker='o', figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._recording = recording
        self._ms = markersize
        self._mark = marker
        self.name = 'ElectrodeGeometry'

    def plot(self, width=4, height=4):
        self._do_plot(width=width, height=height)

    def _do_plot(self, width, height):
        R = self._recording
        geom = np.array(R.get_channel_locations())


        self.ax.axis('off')

        x = geom[:, 0]
        y = geom[:, 1]
        xmin = np.min(x)
        xmax = np.max(x)
        ymin = np.min(y)
        ymax = np.max(y)

        margin = np.maximum(xmax - xmin, ymax - ymin) * 0.2

        self.ax.scatter(x, y, marker=self._mark, s=int(self._ms))
        self.ax.axis('equal')
        self.ax.set_xticks([])
        self.ax.set_yticks([])
        self.ax.set_xlim(xmin - margin, xmax + margin)
        self.ax.set_ylim(ymin - margin, ymax + margin)



import numpy as np
import matplotlib
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseWidget


def plot_multicomp_graph(multi_sorting_comparison, sorter_names=None, draw_labels=False, node_cmap='viridis',
                         edge_cmap='hot_r', alpha_edges=0.7, colorbar=False, figure=None, ax=None):
    """
    Plots multi sorting comparison graph.

    Parameters
    ----------
    multi_sorting_comparison: MultiSortingComparison
        The multi sorting comparison object
    sorter_names: list
        The names of the sorters
    draw_labels: bool
        If True unit labels are shown
    node_cmap: matplotlib colormap
        The colormap to be used for the nodes (default 'viridis')
    edge_cmap: matplotlib colormap
        The colormap to be used for the edges (default 'hot')
    alpha_edges: float
        Alpha value for edges
    colorbar: bool
        If True a colorbar for the edges is plotted
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: MultiCompGraphWidget
        The output widget
    """
    try:
        import networkx as nx
    except ImportError as e:
        raise ImportError('Install networkx to use the multi comparison widget.')
    W = MultiCompGraphWidget(
        multi_sorting_comparison=multi_sorting_comparison,
        sorternames=sorter_names,
        node_cmap=node_cmap,
        edge_cmap=edge_cmap,
        drawlabels=draw_labels,
        alpha_edges=alpha_edges,
        colorbar=colorbar,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class MultiCompGraphWidget(BaseWidget):
    def __init__(self, *, multi_sorting_comparison, sorternames=None, drawlabels=False, node_cmap='viridis',
                 edge_cmap='hot', alpha_edges=0.5, colorbar=False, figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._msc = multi_sorting_comparison
        self._sorter_names = sorternames
        self._drawlabels = drawlabels
        self._node_cmap = node_cmap
        self._edge_cmap = edge_cmap
        self._colorbar = colorbar
        self._alpha_edges = alpha_edges
        self.name = 'MultiCompGraph'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        import networkx as nx

        g = self._msc.graph
        edge_col = []
        for e in g.edges(data=True):
            n1, n2, d = e
            edge_col.append(d['weight'])
        nodes_col = np.array([])
        for i, sort in enumerate(self._msc.get_sorting_list()):
            nodes_col = np.concatenate((nodes_col, np.array([i] * len(sort.get_unit_ids()))))
        nodes_col = nodes_col / len(self._msc.get_sorting_list())

        _ = plt.set_cmap(self._node_cmap)
        _ = nx.draw_networkx_nodes(g, pos=nx.circular_layout(sorted(g)), nodelist=sorted(g.nodes),
                                   node_color=nodes_col, node_size=20, ax=self.ax)
        _ = nx.draw_networkx_edges(g, pos=nx.circular_layout((sorted(g))), nodelist=sorted(g.nodes),
                                   edge_color=edge_col, alpha=self._alpha_edges,
                                   edge_cmap=plt.cm.get_cmap(self._edge_cmap), edge_vmin=self._msc.match_score,
                                   edge_vmax=1, ax=self.ax)
        if self._drawlabels:
            _ = nx.draw_networkx_labels(g, pos=nx.circular_layout((sorted(g))), nodelist=sorted(g.nodes), ax=self.ax)
        if self._colorbar:
            norm = matplotlib.colors.Normalize(vmin=self._msc.match_score, vmax=1)
            cmap = plt.cm.get_cmap(self._edge_cmap)
            m = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
            self.figure.colorbar(m)

        self.ax.axis('off')



import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseMultiWidget


def plot_isi_distribution(sorting, sampling_frequency=None, unit_ids=None, bins=10, window=1, figure=None, ax=None):
    """
    Plots spike train ISI distribution.

    Parameters
    ----------
    sorting: SortingExtractor
        The sorting extractor object
    sampling_frequency: float
        The sampling frequency (if not in the sorting extractor)
    unit_ids: list
        List of unit ids
    bins: int
        Number of bins
    window: float
        Window size in s
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: ISIDistributionWidget
        The output widget
    """
    if sampling_frequency is None:
        if sorting.get_sampling_frequency() is None:
            raise Exception("Sampling rate information is not in the SortingExtractor. "
                            "Provide the 'sampling_frequency' argument")
        else:
            sampling_frequency = sorting.get_sampling_frequency()
    W = ISIDistributionWidget(
        sorting=sorting,
        sampling_frequency=sampling_frequency,
        unit_ids=unit_ids,
        bins=bins,
        window=window,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class ISIDistributionWidget(BaseMultiWidget):
    def __init__(self, *, sorting, sampling_frequency, unit_ids=None, bins=10, window=1, figure=None, ax=None):
        BaseMultiWidget.__init__(self, figure, ax)
        self._sorting = sorting
        self._unit_ids = unit_ids
        self._sampling_frequency = sampling_frequency
        self._bins = bins
        self._maxw = window
        self.name = 'ISIDistribution'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        list_isi = []
        for unit in units:
            times = self._sorting.get_unit_spike_train(unit_id=unit) / float(self._sampling_frequency)
            bin_counts, bin_edges = compute_isi_dist(times, bins=self._bins, maxwindow=self._maxw)
            item = dict(
                title='Unit {}'.format(int(unit)),
                bin_counts=bin_counts,
                bin_edges=bin_edges,
                xticks=[0, self._maxw / 2, self._maxw]
            )
            list_isi.append(item)
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_isi_multi(list_isi)

    def _plot_isi_multi(self, list_isi, *, ncols=5, **kwargs):
        if len(list_isi) < ncols:
            ncols = len(list_isi)
        nrows = np.ceil(len(list_isi) / ncols)
        for i, item in enumerate(list_isi):
            ax = self.get_tiled_ax(i, nrows, ncols, hspace=0.7)
            _plot_isi(**item, **kwargs, ax=ax)


def _plot_isi( *, bin_counts, bin_edges, ax, xticks=None, title=''):
    bins = bin_edges[:-1] + np.mean(np.diff(bin_edges))
    wid = np.mean(np.diff(bins))  #(bin_edges[1] - bin_edges[0]) * len(bin_edges)
    ax.bar(x=bins, height=bin_counts, width=wid, color='gray', align='edge')
    if xticks is not None:
        ax.set_xticks(xticks)
    ax.set_xlabel('dt (s)')
    ax.set_yticks([])
    if title:
        ax.set_title(title, color='gray')


def compute_isi_dist(times, *, bins, maxwindow=10):
    isi = np.diff(times)
    isi = isi[isi < maxwindow]
    # bins = np.arange(0, maxisi, binsize)
    bin_counts, bin_edges = np.histogram(isi, bins=bins, density=True)
    return bin_counts, bin_edges



from spikewidgets.widgets.basewidget import BaseWidget
import spikecomparison as sc


def plot_sorting_performance(gt_sorting_comparison, property_name=None, metric='accuracy', markersize=10, marker='.',
                             figure=None, ax=None):
    """
    Plots sorting performance for each ground-truth unit.

    Parameters
    ----------
    gt_sorting_comparison: GroundTruthComparison
        The ground truth sorting comparison object
    property_name: str
        The property of the sorting extractor to use as x-axis (e.g. snr).
        If None, no property is used.
    metric: str
        The performance metric. 'accuracy' (default), 'precision', 'recall', 'miss rate', etc.
    markersize: int
        The size of the marker
    marker: str
        The matplotlib marker to use (default '.')
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: SortingPerformanceWidget
        The output widget
    """
    W = SortingPerformanceWidget(
        sorting_comparison=gt_sorting_comparison,
        property_name=property_name,
        figure=figure,
        markersize=markersize,
        marker=marker,
        metric=metric,
        ax=ax
    )
    W.plot()
    return W


class SortingPerformanceWidget(BaseWidget):
    def __init__(self, *, sorting_comparison, property_name=None, metric='accuracy', markersize=10,
                 marker='.', figure=None, ax=None):
        assert isinstance(sorting_comparison, sc.GroundTruthComparison), \
            "The 'sorting_comparison' object should be a GroundTruthComparison instance"
        BaseWidget.__init__(self, figure, ax)
        self._SC = sorting_comparison
        self._property_name = property_name
        self._metric = metric
        self._ms = markersize
        self._mark = marker
        self.name = 'SortingPerformance'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        SC = self._SC
        units = SC.sorting1.get_unit_ids()
        perf = SC.get_performance()[self._metric]
        if self._property_name is not None:
            assert self._property_name in SC.sorting1.get_shared_unit_property_names(), "%s should be " \
                                                                                 "a property of the ground truth " \
                                                                                 "sorting extractor"
            xvals = SC.sorting1.get_units_property(unit_ids=units, property_name=self._property_name)
            self.ax.plot(xvals, perf, marker=self._mark, markersize=int(self._ms), ls='')
            self.ax.set_xlabel(self._property_name)
        else:
            self.ax.plot(perf, '.')
            self.ax.set_xticks([])
        self.ax.set_ylabel(self._metric)
        self.ax.set_ylim([-0.05, 1.05])



from .widgets.timeserieswidget import *
from .widgets.unitwaveformswidget import *
from .widgets.correlogramswidget import *
from .widgets.sortingperformancewidget import *
from .widgets.electrodegeometrywidget import *
from .widgets.confusionmatrixwidget import *
from .widgets.rasterswidget import *
from .widgets.isidistwidget import *
from .widgets.multicompgraphwidget import *
from .widgets.amplitudewidget import *
from .widgets.featurewidgets import *
from .widgets.spectrumwidget import *
from .widgets.agreementmatrixwidget import *

from .version import version as __version__



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
from spikewidgets.widgets.basewidget import BaseWidget


def plot_timeseries(recording, channel_ids=None, trange=None, color_groups=False,
                    figure=None, ax=None):
    """
    Plots electrode geometry.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    channel_ids: list
        The channel ids to display.
    trange: list
        List with start time and end time
    color_groups: bool
        If True groups are plotted with different colors
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: TimeseriesWidget
        The output widget
    """
    W = TimeseriesWidget(
        recording=recording,
        channel_ids=channel_ids,
        trange=trange,
        color_groups=color_groups,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class TimeseriesWidget(BaseWidget):
    def __init__(self, *, recording, channel_ids=None, trange=None,
                 color_groups=False, figure=None,  ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._recording = recording
        self._sampling_frequency = recording.get_sampling_frequency()
        self._visible_channels = channel_ids
        if self._visible_channels is None:
            self._visible_channels = recording.get_channel_ids()
        self._visible_trange = trange
        if self._visible_trange is None:
            self._visible_trange = [0, np.minimum(10000, recording.get_num_frames())]
        else:
            assert len(trange) == 2, "'trange' should be a list with start and end time in seconds"
            self._visible_trange = [int(t * recording.get_sampling_frequency()) for t in trange]
        self._initialize_stats()
        # self._vspacing = self._mean_channel_std * 20
        self._vspacing = self._max_channel_amp * 1.5
        self._visible_trange = self._fix_trange(self._visible_trange)
        self._ax = ax
        self._color_groups = color_groups
        if color_groups:
            self._colors = []
            self._group_color_map = {}
            all_groups = recording.get_channel_groups()
            groups = np.unique(all_groups)
            N = len(groups)
            import colorsys
            HSV_tuples = [(x*1.0/N, 0.5, 0.5) for x in range(N)]
            self._colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples))
            color_idx = 0
            for group in groups:
                self._group_color_map[group] = color_idx
                color_idx += 1
        self.name = 'TimeSeries'
            
    def plot(self):
        self._do_plot()

    def _do_plot(self):
        chunk0 = self._recording.get_traces(
            channel_ids=self._visible_channels,
            start_frame=self._visible_trange[0],
            end_frame=self._visible_trange[1]
        )

        self.ax.set_xlim(self._visible_trange[0] / self._sampling_frequency, self._visible_trange[1] / self._sampling_frequency)
        self.ax.set_ylim(-self._vspacing, self._vspacing * len(self._visible_channels))
        self.ax.get_xaxis().set_major_locator(MaxNLocator(prune='both'))
        self.ax.get_yaxis().set_ticks([])
        self.ax.set_xlabel('time (s)')

        self._plots = {}
        self._plot_offsets = {}
        offset0 = self._vspacing * (len(self._visible_channels) - 1)
        tt = np.arange(self._visible_trange[0], self._visible_trange[1]) / self._sampling_frequency
        for im, m in enumerate(self._visible_channels):
            self._plot_offsets[m] = offset0
            if self._color_groups:
                group = self._recording.get_channel_groups(channel_ids=[m])[0]
                group_color_idx = self._group_color_map[group]
                self._plots[m] = self.ax.plot(tt, self._plot_offsets[m] + chunk0[im, :],
                                              color=self._colors[group_color_idx])
            else:
                self._plots[m] = self.ax.plot(tt, self._plot_offsets[m] + chunk0[im, :])
            offset0 = offset0 - self._vspacing

    def _fix_trange(self, trange):
        N = self._recording.get_num_frames()
        if trange[1] > N:
            # trange[0] += N - trange[1]
            # trange[1] = N - trange[1]
            trange[1] = N
        if trange[0] < 0:
            # trange[1] += -trange[0]
            trange[0] = 0
        # trange[0] = np.maximum(0, trange[0])
        # trange[1] = np.minimum(N, trange[1])
        return trange

    def _initialize_stats(self):
        self._channel_stats = {}
        # M=self._reader.numChannels()
        # N=self._reader.numTimepoints()
        chunk0 = self._recording.get_traces(
            channel_ids=self._visible_channels,
            start_frame=self._visible_trange[0],
            end_frame=self._visible_trange[1]
        )
        # chunk0=self._reader.getChunk(channels=self._visible_channels,trange=self._visible_trange)
        M0 = chunk0.shape[0]
        for ii in range(M0):
            self._channel_stats[self._visible_channels[ii]] = _compute_channel_stats_from_data(chunk0[ii, :])
        self._mean_channel_std = np.mean([self._channel_stats[m]['std'] for m in self._visible_channels])
        self._max_channel_amp = np.max([self._channel_stats[m]['max_amp'] for m in self._visible_channels])


def _compute_channel_stats_from_data(X):
    return dict(
        mean=np.mean(X),
        std=np.std(X),
        max_amp=np.max(np.abs(X))
    )



import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseWidget


def plot_confusion_matrix(gt_comparison, count_text=True, unit_ticks=True,
                          ax=None, figure=None):
    """
    Plots sorting comparison confusion matrix.

    Parameters
    ----------
    gt_comparison: GroundTruthComparison
        The ground truth sorting comparison object
    count_text: bool
        If True counts are displayed as text
    unit_ticks: bool
        If True unit tick labels are displayed
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: ConfusionMatrixWidget
        The output widget
    """
    W = ConfusionMatrixWidget(
        gt_comparison=gt_comparison,
        count_text=count_text,
        unit_ticks=unit_ticks,
        figure=figure,
        ax=ax,
    )
    W.plot()
    return W


class ConfusionMatrixWidget(BaseWidget):
    def __init__(self, *, gt_comparison, count_text=True, unit_ticks=True,
                 figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._gtcomp = gt_comparison
        self._count_text = count_text
        self._unit_ticks = unit_ticks
        self.name = 'ConfusionMatrix'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        # a dataframe
        confusion_matrix = self._gtcomp.get_confusion_matrix()

        N1 = confusion_matrix.shape[0] - 1
        N2 = confusion_matrix.shape[1] - 1

        # Using matshow here just because it sets the ticks up nicely. imshow is faster.
        self.ax.matshow(confusion_matrix.values, cmap='Greens')

        if self._count_text:
            for (i, j), z in np.ndenumerate(confusion_matrix.values):
                if z != 0:
                    if z > np.max(confusion_matrix.values) / 2.:
                        self.ax.text(j, i, '{:d}'.format(z), ha='center', va='center', color='white')
                    else:
                        self.ax.text(j, i, '{:d}'.format(z), ha='center', va='center', color='black')

        self.ax.axhline(int(N1 - 1) + 0.5, color='black')
        self.ax.axvline(int(N2 - 1) + 0.5, color='black')

        # Major ticks
        self.ax.set_xticks(np.arange(0, N2 + 1))
        self.ax.set_yticks(np.arange(0, N1 + 1))
        self.ax.xaxis.tick_bottom()
        
        # Labels for major ticks
        if self._unit_ticks:
            self.ax.set_yticklabels(confusion_matrix.index, fontsize=12)
            self.ax.set_xticklabels(confusion_matrix.columns, fontsize=12)
        else:
            self.ax.set_xticklabels(np.append([''] * N2, 'FN'), fontsize=10)
            self.ax.set_yticklabels(np.append([''] * N1, 'FP'), fontsize=10)

        self.ax.set_xlabel(self._gtcomp.name_list[1], fontsize=20)
        self.ax.set_ylabel(self._gtcomp.name_list[0], fontsize=20)
        
        self.ax.set_xlim(-0.5, N2+0.5)
        self.ax.set_ylim(N1+0.5, -0.5, )
        
        




import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseWidget


def plot_rasters(sorting, sampling_frequency=None, unit_ids=None, trange=None, color='k', figure=None, ax=None):
    """
    Plots spike train rasters.

    Parameters
    ----------
    sorting: SortingExtractor
        The sorting extractor object
    sampling_frequency: float
        The sampling frequency (if not in the sorting extractor)
    unit_ids: list
        List of unit ids
    trange: list
        List with start time and end time
    color: matplotlib color
        The color to be used
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: RasterWidget
        The output widget
    """
    if sampling_frequency is None:
        if sorting.get_sampling_frequency() is None:
            raise Exception("Sampling rate information is not in the SortingExtractor. "
                            "Provide the 'sampling_frequency' argument")
        else:
            sampling_frequency = sorting.get_sampling_frequency()
    W = RasterWidget(
        sorting=sorting,
        sampling_frequency=sampling_frequency,
        unit_ids=unit_ids,
        trange=trange,
        color=color,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class RasterWidget(BaseWidget):
    def __init__(self, *, sorting, sampling_frequency, unit_ids=None, trange=None, color='k', figure=None, ax=None):
        BaseWidget.__init__(self, figure, ax)
        self._SX = sorting
        self._unit_ids = unit_ids
        self._figure = None
        self._sampling_frequency = sampling_frequency
        self._color = color
        self._max_frame = 0
        for unit_id in self._SX.get_unit_ids():
            spike_train = self._SX.get_unit_spike_train(unit_id)
            curr_max_frame = np.max(spike_train)
            if curr_max_frame > self._max_frame:
                self._max_frame = curr_max_frame
        self._visible_trange = trange
        if self._visible_trange is None:
            self._visible_trange = [0, self._max_frame]
        else:
            assert len(trange) == 2, "'trange' should be a list with start and end time in seconds"
            self._visible_trange = [int(t * sampling_frequency) for t in trange]
        self._visible_trange = self._fix_trange(self._visible_trange)
        self.name = 'Raster'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        units_ids = self._unit_ids
        if units_ids is None:
            units_ids = self._SX.get_unit_ids()

        with plt.rc_context({'axes.edgecolor': 'gray'}):
            for u_i, unit_id in enumerate(units_ids):
                t = self._SX.get_unit_spike_train(unit_id, start_frame=self._visible_trange[0], end_frame=self._visible_trange[1]) / float(self._sampling_frequency)
                self.ax.plot(t, u_i * np.ones_like(t), marker='|', mew=1, markersize=3,
                             ls='', color=self._color)
            visible_start_frame = self._visible_trange[0] / float(self._sampling_frequency)
            visible_end_frame = self._visible_trange[1] / float(self._sampling_frequency)
            self.ax.set_yticks(np.arange(len(units_ids)))
            self.ax.set_yticklabels(units_ids)
            self.ax.set_xlim(visible_start_frame, visible_end_frame)
            self.ax.set_xlabel('time (s)')

    def _fix_trange(self, trange):
        if trange[1] > self._max_frame:
            # trange[0] += max_t - trange[1]
            trange[1] = self._max_frame
        if trange[0] < 0:
            # trange[1] += -trange[0]
            trange[0] = 0
        # trange[0] = np.maximum(0, trange[0])
        # trange[1] = np.minimum(max_t, trange[1])
        return trange



import setuptools

d = {}
exec(open("spikewidgets/version.py").read(), None, d)
version = d['version']
pkg_name = "spikewidgets"
long_description = open("README.md").read()

setuptools.setup(
    name=pkg_name,
    version=version,
    author="Jeremy Magland, Alessio Buccino",
    author_email="jmagland@flatironinstitute.org",
    description="Python widgets for spike sorting and electrophysiology",
    url="https://github.com/SpikeInterface/spikewidgets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'spiketoolkit',
        'spikecomparison',
        'matplotlib',
        'MEAutility>=1.4.3'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



import numpy as np
from matplotlib import pyplot as plt
import spiketoolkit as st
import MEAutility as mu
from spikewidgets.widgets.basewidget import BaseMultiWidget


def plot_unit_waveforms(recording, sorting, channel_ids=None, unit_ids=None, ms_before=1., ms_after=2.,
                        max_spikes_per_unit=100, max_channels=16, channel_locs=True,
                        plot_templates=True, figure=None, ax=None):
    """
    Plots unit waveforms.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    sorting: SortingExtractor
        The sorting extractor object
    channel_ids: list
        The channel ids to display.
    unit_ids: list
        List of unit ids.
    ms_before: float
        Time before peak (ms)
    ms_after: float
        Time after peak (ms)
    max_spikes_per_unit: int
        Maximum number of spikes to display per unit.
    max_channels: int
        Maximum number of largest channels to plot waveform
    channel_locs: bool
        If True, channel locations are used to display the waveforms.
        If False, waveforms are displayed in vertical order. (default)
    plot_templates: bool
        If True, templates are plotted over the waveforms
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: UnitWaveformsWidget
        The output widget
    """
    W = UnitWaveformsWidget(
        recording=recording,
        sorting=sorting,
        channel_ids=channel_ids,
        unit_ids=unit_ids,
        max_spikes_per_unit=max_spikes_per_unit,
        max_channels=max_channels,
        ms_before=ms_before,
        ms_after=ms_after,
        channel_locs=channel_locs,
        plot_templates=plot_templates,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


def plot_unit_templates(recording, sorting, channel_ids=None, unit_ids=None, ms_before=1., ms_after=2.,
                        max_spikes_per_unit=100, max_channels=16, channel_locs=True,
                        figure=None, ax=None):
    """
    Plots unit waveforms.

    Parameters
    ----------
    recording: RecordingExtractor
        The recordng extractor object
    sorting: SortingExtractor
        The sorting extractor object
    channel_ids: list
        The channel ids to display.
    unit_ids: list
        List of unit ids.
    ms_before: float
        Time before peak (ms)
    ms_after: float
        Time after peak (ms)
    max_spikes_per_unit: int
        Maximum number of spikes to display per unit.
    max_channels: int
        Maximum number of largest channels to plot waveform
    channel_locs: bool
        If True, channel locations are used to display the waveforms.
        If False, waveforms are displayed in vertical order. (default)
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: UnitWaveformsWidget
        The output widget
    """
    W = UnitTemplatesWidget(
        recording=recording,
        sorting=sorting,
        channel_ids=channel_ids,
        unit_ids=unit_ids,
        max_spikes_per_unit=max_spikes_per_unit,
        max_channels=max_channels,
        ms_before=ms_before,
        ms_after=ms_after,
        channel_locs=channel_locs,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class UnitWaveformsWidget(BaseMultiWidget):
    def __init__(self, *, recording, sorting, channel_ids=None, unit_ids=None, max_spikes_per_unit=50,
                 max_channels=16, ms_before=1., ms_after=2., channel_locs=True, plot_templates=True,
                 figure=None, ax=None):
        BaseMultiWidget.__init__(self, figure, ax)
        self._recording = recording
        self._sorting = sorting
        self._channel_ids = channel_ids
        self._max_channels = max_channels
        self._unit_ids = unit_ids
        self._ms_before = ms_before
        self._ms_after = ms_after
        self._max_spikes_per_unit = max_spikes_per_unit
        self._ch_locs = channel_locs
        self.name = 'UnitWaveforms'
        self._plot_waveforms = True
        self._plot_templates = plot_templates

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        unit_ids = self._unit_ids
        channel_ids = self._channel_ids
        if unit_ids is None:
            unit_ids = self._sorting.get_unit_ids()
        if channel_ids is None:
            channel_ids = self._recording.get_channel_ids()
        if 'location' in self._recording.get_shared_channel_property_names():
            all_locations = np.array(self._recording.get_channel_locations())
            channel_locations = np.array(self._recording.get_channel_locations(channel_ids=channel_ids))
        else:
            all_locations = None
            channel_locations = None
        list_spikes = []
        max_channels_list = []
        for unit_id in unit_ids:
            spiketrain = self._sorting.get_unit_spike_train(unit_id=unit_id)
            if spiketrain is not None:
                random_wf = st.postprocessing.get_unit_waveforms(recording=self._recording, sorting=self._sorting,
                                                                 unit_ids=[unit_id], channel_ids=channel_ids,
                                                                 ms_before=self._ms_before, ms_after=self._ms_after,
                                                                 max_spikes_per_unit=self._max_spikes_per_unit,
                                                                 save_as_features=False)[0]
                if self._max_channels < self._recording.get_num_channels():
                    template = np.mean(random_wf, axis=0)
                    peak_idx = np.unravel_index(np.argmax(np.abs(template)),
                                                template.shape)[1]
                    max_channel_idxs = np.argsort(np.abs(template[:, peak_idx]))[::-1][:self._max_channels]
                    max_channels_list.append(max_channel_idxs)
                else:
                    max_channels_list.append(np.arange(len(channel_ids)))
                spikes = random_wf
                item = dict(
                    representative_waveforms=spikes,
                    title='Unit {}'.format(int(unit_id))
                )
                list_spikes.append(item)
            else:
                print(unit_id, ' spike train is None')

        with plt.rc_context({'axes.edgecolor': 'gray'}):
            if self._ch_locs:
                self._plot_spike_shapes_multi(list_spikes, channel_locations=channel_locations,
                                              all_locations=all_locations, max_channels_list=max_channels_list,
                                              plot_templates=self._plot_templates, plot_waveforms=self._plot_waveforms)
            else:
                self._plot_spike_shapes_multi(list_spikes, channel_locations=None, max_channels_list=max_channels_list,
                                              plot_templates=self._plot_templates, plot_waveforms=self._plot_waveforms)

    def _plot_spike_shapes_multi(self, list_spikes, max_channels_list, *, ncols=5,  **kwargs):
        vscale, ylim = _determine_global_vscale_ylim(list_spikes)
        if len(list_spikes) < ncols:
            ncols = len(list_spikes)
        nrows = np.ceil(len(list_spikes) / ncols)
        for i, item in enumerate(list_spikes):
            ax = self.get_tiled_ax(i, nrows, ncols)
            _plot_spike_shapes(**item, **kwargs, ax=ax, channels=max_channels_list[i], vscale=vscale, ylim_wf=ylim)


class UnitTemplatesWidget(UnitWaveformsWidget):
    def __init__(self, *, recording, sorting, channel_ids=None, unit_ids=None, max_spikes_per_unit=50,
                 max_channels=16, ms_before=1., ms_after=2., channel_locs=True, figure=None, ax=None):
        UnitWaveformsWidget.__init__(self, recording=recording, sorting=sorting, channel_ids=channel_ids,
                                     unit_ids=unit_ids, max_spikes_per_unit=max_spikes_per_unit,
                                     max_channels=max_channels, ms_before=ms_before, ms_after=ms_after,
                                     channel_locs=channel_locs, plot_templates=True, figure=figure, ax=ax)
        self.name = 'UnitTemplates'
        self._plot_waveforms = False


def _plot_spike_shapes(*, ax, channels, representative_waveforms=None, channel_locations=None,
                       all_locations=None, color='blue', vscale=None, plot_waveforms=True, plot_templates=True,
                       ylim_wf=None, title=''):
    if representative_waveforms is None:
        raise Exception('You must provide either average_waveform, representative waveforms, or both')
    waveforms = representative_waveforms
    average_waveform = np.mean(waveforms, axis=0)

    M = average_waveform.shape[0]  # number of channels
    if channel_locations is None:
        channel_locations = np.zeros((M, 2))
        for m in range(M):
            channel_locations[m, :] = [0, -m]
        all_locations = channel_locations
    else:
        if all_locations is not None:
            if all_locations.shape == channel_locations.shape:
                all_locations = channel_locations

    probe = mu.return_mea(info={'pos': all_locations})

    if vscale is None:
        vscale = 1.5 * np.max(np.abs(average_waveform))

    ylim_g = np.array([np.min(all_locations[:, 1]), np.max(all_locations[:, 1])])
    ylim_g *= 1.05

    # fix for linear horizontal layout
    if ylim_g[0] == ylim_g[1]:
        ylim = ylim_wf / vscale
    else:
        ylim = ylim_g

    if plot_waveforms:
        ax = mu.plot_mea_recording(waveforms, probe, colors=(0.5, 0.5, 0.5), alpha=0.3, lw=0.3, ax=ax, vscale=vscale,
                                   ylim=ylim, channels=channels)
    if plot_templates:
        ax = mu.plot_mea_recording(average_waveform, probe, colors=color, ax=ax, lw=2, vscale=vscale, ylim=ylim,
                                   channels=channels)

    ax.set_xticks([])
    ax.set_yticks([])
    if title:
        ax.set_title(title, color='gray')


def _get_vscale_ylim_for_item(average_waveform=None, representative_waveforms=None):
    if average_waveform is None:
        if representative_waveforms is None:
            raise Exception('You must provide either average_waveform, representative waveforms, or both')
        average_waveform = np.mean(representative_waveforms, axis=0)
    vscale = np.max(np.abs(average_waveform))
    ylim = [np.min(average_waveform), np.max(average_waveform)]
    return vscale, ylim


def _determine_global_vscale_ylim(list_spikes):
    vscales = []
    ylims = []
    for item in list_spikes:
        vscale, ylim = _get_vscale_ylim_for_item(
            average_waveform=item.get('average_waveform', None),
            representative_waveforms=item.get('representative_waveforms', None)
        )
        vscales.append(vscale)
        ylims.append(ylim)
    vscale_global = np.median(vscales) * 1.1
    ylims = np.array(ylims)
    ylim_global = [np.min(ylims[:, 0]), np.max(ylims[:, 1])]
    ylim_global = [ylim_global[0] - 0.1*np.ptp(ylim_global), ylim_global[1] + 0.1*np.ptp(ylim_global)]
    return vscale_global, ylim_global



import numpy as np
from matplotlib import pyplot as plt
from spikewidgets.widgets.basewidget import BaseMultiWidget
from .correlograms_phy import compute_correlograms


def plot_autocorrelograms(sorting, sampling_frequency=None, unit_ids=None, bin_size=2, window=50,
                          figure=None, ax=None):
    """
    Plots spike train auto-correlograms.

    Parameters
    ----------
    sorting: SortingExtractor
        The sorting extractor object
    sampling_frequency: float
        The sampling frequency (if not in the sorting extractor)
    unit_ids: list
        List of unit ids
    bin_size: float
        Bin size in s
    window: float
        Window size in s
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: AutoCorrelogramsWidget
        The output widget
    """
    if sampling_frequency is None:
        if sorting.get_sampling_frequency() is None:
            raise Exception("Sampling rate information is not in the SortingExtractor. "
                            "Provide the 'sampling_frequency' argument")
        else:
            sampling_frequency = sorting.get_sampling_frequency()
    W = AutoCorrelogramsWidget(
        sorting=sorting,
        sampling_frequency=sampling_frequency,
        unit_ids=unit_ids,
        binsize=bin_size,
        window=window,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


def plot_crosscorrelograms(sorting, sampling_frequency=None, unit_ids=None, bin_size=1, window=10,
                           figure=None, ax=None):
    """
    Plots spike train cross-correlograms.

    Parameters
    ----------
    sorting: SortingExtractor
        The sorting extractor object
    sampling_frequency: float
        The sampling frequency (if not in the sorting extractor)
    unit_ids: list
        List of unit ids
    bin_size: float
        Bin size in s
    window: float
        Window size in s
    figure: matplotlib figure
        The figure to be used. If not given a figure is created
    ax: matplotlib axis
        The axis to be used. If not given an axis is created

    Returns
    -------
    W: CrossCorrelogramsWidget
        The output widget
    """
    if sampling_frequency is None:
        if sorting.get_sampling_frequency() is None:
            raise Exception("Sampling rate information is not in the SortingExtractor. "
                            "Provide the 'sampling_frequency' argument")
        else:
            sampling_frequency = sorting.get_sampling_frequency()
    W = CrossCorrelogramsWidget(
        sorting=sorting,
        sampling_frequency=sampling_frequency,
        unit_ids=unit_ids,
        binsize=bin_size,
        window=window,
        figure=figure,
        ax=ax
    )
    W.plot()
    return W


class AutoCorrelogramsWidget(BaseMultiWidget):
    def __init__(self, *, sorting, sampling_frequency, unit_ids=None, binsize=2, window=50, figure=None, ax=None):
        BaseMultiWidget.__init__(self, figure, ax)
        self._sorting = sorting
        self._unit_ids = unit_ids
        self._sampling_frequency = sampling_frequency
        self._binsize = binsize
        self._window = window
        self.name = 'AutoCorrelograms'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        list_corr = []

        spike_times, spike_clusters = _prepare_spike_times_and_clusters(self._sorting, units, self._sampling_frequency)
        ccg = compute_correlograms(spike_times, spike_clusters, cluster_ids=units, sample_rate=self._sampling_frequency,
                                   bin_size=self._binsize, window_size=self._window, symmetrize=True)

        for i_u, unit in enumerate(units):
            bin_counts = ccg[i_u, i_u]
            bins = np.linspace(- self._window / 2, self._window / 2, len(bin_counts))
            wid = self._binsize
            item = dict(
                title='Unit {}'.format(int(unit)),
                bin_counts=bin_counts,
                bins=bins,
                wid=wid
            )
            list_corr.append(item)

        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_autocorrelograms_multi(list_corr)

    def _plot_autocorrelograms_multi(self, list_corr, *, ncols=5, **kwargs):
        if len(list_corr) < ncols:
            ncols = len(list_corr)
        nrows = np.ceil(len(list_corr) / ncols)
        for i, item in enumerate(list_corr):
            ax = self.get_tiled_ax(i, nrows, ncols, hspace=0.7)
            _plot_correlogram(**item, **kwargs, ax=ax, color='gray')
            

class CrossCorrelogramsWidget(BaseMultiWidget):
    def __init__(self, *, sorting, sampling_frequency, unit_ids=None, binsize=2, window=50, figure=None, ax=None):
        BaseMultiWidget.__init__(self, figure, ax)
        self._sorting = sorting
        self._unit_ids = unit_ids
        self._sampling_frequency = sampling_frequency
        self._binsize = binsize
        self._window = window
        self.name = 'CrossCorrelograms'

    def plot(self):
        self._do_plot()

    def _do_plot(self):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        list_corr = []

        spike_times, spike_clusters = _prepare_spike_times_and_clusters(self._sorting, units, self._sampling_frequency)
        ccg = compute_correlograms(spike_times, spike_clusters, cluster_ids=units, sample_rate=self._sampling_frequency,
                                   bin_size=self._binsize, window_size=self._window, symmetrize=True)

        for u1, unit1 in enumerate(units):
            for u2, unit2 in enumerate(units):
                bin_counts = ccg[u1, u2]
                bins = np.linspace(- self._window / 2, self._window / 2, len(bin_counts))
                wid = self._binsize
                if u1 == u2:
                    title = 'Unit {}'.format(int(unit1))
                else:
                    title = 'Units {}-{}'.format(int(unit1), int(unit2))
                item = dict(
                    title=title,
                    bin_counts=bin_counts,
                    bins=bins,
                    wid=wid
                )
                list_corr.append(item)
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_crosscorrelograms_multi(list_corr)

    def _plot_crosscorrelograms_multi(self, list_corr, **kwargs):
        units = self._unit_ids
        if units is None:
            units = self._sorting.get_unit_ids()
        ncols = len(units)
        nrows = np.ceil(len(list_corr) / ncols)
        self.figure.set_size_inches((3*ncols, 2*nrows))
        for i, item in enumerate(list_corr):
            ax, diag = self.get_tiled_ax(i, nrows, ncols, hspace=1.5, wspace=0.2, is_diag=True)
            if diag:
                _plot_correlogram(**item, **kwargs, ax=ax, color='blue')
            else:
                _plot_correlogram(**item, **kwargs, ax=ax, color='gray')


def _plot_correlogram(*, ax, bin_counts, bins, wid, title='', color=None):
    ax.bar(x=bins, height=bin_counts, width=wid, color=color, align='edge')
    ax.set_xlabel('dt (s)')
    ax.set_xticks([bins[0], bins[len(bins)//2], bins[-1]])
    ax.set_yticks([])
    if title:
        ax.set_title(title, color='gray')


def _prepare_spike_times_and_clusters(sorting, unit_ids, sampling_frequency):
    spike_times = np.array([])
    spike_clusters = np.array([], dtype=int)

    for u in sorting.get_unit_ids():
        if u in unit_ids:
            spike_times = np.concatenate((spike_times, sorting.get_unit_spike_train(unit_id=u) / sampling_frequency))
            spike_clusters = np.concatenate((spike_clusters, np.array([u]*len(sorting.get_unit_spike_train(unit_id=u)))))

    order = np.argsort(spike_times)
    spike_times = spike_times[order]
    spike_clusters = spike_clusters[order]

    return spike_times, spike_clusters
